/*
 * Copyright (C) 2005 Bernhard Erb

 *This program is free software; you can redistribute it and/or
 *modify it under the terms of the GNU General Public License
 *as published by the Free Software Foundation; either version 2
 *of the License, or (at your option) any later version.

 *This program is distributed in the hope that it will be useful,
 *but WITHOUT ANY WARRANTY; without even the implied warranty of
 *MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *GNU General Public License for more details.

 *You should have received a copy of the GNU General Public License
 *along with this program; if not, write to the Free Software
 *Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

/************************************************************
 * project=tweety                                           *
 * codename=kleiner eibnet/ip server                        *
 * version=1.1                                              *
 ************************************************************/

/**
 * @file send_udp.c
 *
 * This file contains the code for the UDP send process.
 *
 * @date created 16.05.05,
 *       last modified 02.09.05
 *
 * @author bernhard erb
 *         bernhard.erb@gmx.net
 *
 */ 

/*
 * ********************************************* includes ***
 */


#include "common.h"


/*
 * ********************************************** globals ***
 */


int send_udp_pipe = -1;
int ntimeout;
byte *buffer;
int cemi_length;
CEMI_L_DATA_MESSAGE *cemi;


/*
 * ******************************************* prototypes ***
 */


void SendUdp_HandleSignal(int sig);	/* forward declaration */
void SendUdp_SendPacket();	/* forward declaration */
void readNextFromPipe();	/* forward declaration */
void bailOut(char *message);	/* forward declaration: method in header.c */
void sendPacket(EIBNETIP_PACKET * packet, struct sockaddr_in address);	/* forward declaration: method in header.c */
void *sendDisconnect_Request();	/* forward declaration: method in receive_udp.c */

/*
 * ******************************************** functions ***
 */

/**
 * This method is the main Send_UDP
 * subprocess main method.
 * It allocates all for the process
 * relevant ressources:
 *
 * - registers all needed signals
 * - ataches shared memory
 * - opens the semaphor
 * - inits the sendpipe
 * and then calls the readNextFromPipe
 * method.
 */
void sendUdp(void)
{

    /*
     * *** register signals ***
     */

    signal(SIGINT, SendUdp_HandleSignal);	/* terminating signal */
    signal(SIGALRM, SendUdp_HandleSignal);	/* alarm signal */
    signal(SIGUSR1, SendUdp_HandleSignal);	/* signal for tunnelling ack */

    /*
     * *** attach shared memory ***
     */

    if ((active_connection = (connection *) shmat(nSharedMem,
						  (const void *) 0,
						  0)) ==
	(connection *) - 1) {
	bailOut("Error: can't attach memory send udp");
    }

    active_connection->reset = 1;	/* reset active connection */

    /*
     * *** open semaphore ***
     */

    if ((semid = sem_open(KEY_SEMAPHOR, 0, PERM, 2)) == (sem_t *) NULL) {
	bailOut("Error: can't grab semaphor send udp");
    }

    /*
     * *** init pipe *** 
     */

    if ((send_udp_pipe = open(SEND_PIPE, O_RDONLY)) == -1) {	/* init pipe */
	fprintf(stdout, "%s\n", SEND_PIPE);
	bailOut("Error: can't open pipe send_pipe");
    }

    buffer = (byte *) malloc(128);	/* init buffer */
    readNextFromPipe();		/* read next message from pipe */

}

/**
 * closes the send pipe
 */
void SendUdp_FreeRessources()
{

    if ((send_udp_pipe != -1)) {	/* close pipe */
	close(send_udp_pipe);
    }

}

/**
 * This method is the signal
 * handler for the subsequent
 * signals:
 * SIGUSR1: this signals i sent
 * by the receive_udp process,
 * communicating the reception of
 * a tunnelling acknowledge frame.
 * SIGALRM: This signal acts as 
 * sending timeout. The frame is the
 * processed once again, if no ack is
 * detected the clinet is disconnected.
 * SIGINT: With the reception of 
 * this signal the process is halted.
 */
void SendUdp_HandleSignal(int sig)
{
    switch (sig) {
    case SIGUSR1:		/* Tunnelling ack received */
	{
	    alarm(0);		/* reset alarm */
	    ntimeout = 0;	/* reset number of timeouts */
	    break;
	}
    case SIGALRM:		/* Timeout reachead */
	{
	    ntimeout++;
	    if (ntimeout == 1) {	/* It's the first try again */
		SendUdp_SendPacket();
	    } else if (ntimeout > 1) {	/* The 2 timeout: close connection */
#ifdef DEBUG
		fprintf(stdout,
			"DEBUG: Timeout error connection closed \n");
#endif
		sendDisconnect_Request();	/* disconnect client */
	    }
	    break;
	}
    case SIGINT:		/* end process */
	{
	    SendUdp_FreeRessources();	/* cleanup */
	    exit(0);		/* exit with error code 0 */
	    break;
	}
    default:			/* never reached */
	{
#ifdef DEBUG
	    fprintf(stdout, "DEBUG:Not known signal \n");
#endif
	    break;
	}
    }
}

/**
 * This method converts the BCU 
 * EMI format into the medium 
 * indipendent cEMI message format
 * used by EIBnet/IP
 */
byte *convert_EMI_to_CEMI(byte * buffer)
{
    /* emi->ctrl2&0x0f is the data length */
    EMI_L_DATA_MESSAGE *emi = (EMI_L_DATA_MESSAGE *) buffer;
    cemi = malloc(sizeof(CEMI_L_DATA_MESSAGE) + (emi->ctrl2 & 0x0f));

    /* accept only L_DATA_ frames */
    if (emi->mc == 0x11 || emi->mc == 0x49 || emi->mc == 0x4E) {
	switch (emi->mc) {	/* set messagecode */
	case 0x11:
	    cemi->mc = 0x11;
	    break;
	case 0x49:
	    cemi->mc = 0x29;
	    break;
	case 0x4E:
	    cemi->mc = 0x2E;
	    break;
	}

	cemi->addil = 0;	/* only standardmessages are accepted */
	cemi->ctrl1 = emi->ctrl1;
	cemi->saddr = *(word *) & (emi->saddr);
	cemi->daddr = *(word *) & (emi->daddr);
	cemi->ctrl2 = emi->ctrl2 & 0xF0;
	cemi->apci_length = emi->ctrl2 & 0x0F;
	cemi->apci = emi->npdu;
	cemi_length = cemi->apci_length;

	/* copy the data content of the packet */
	memcpy(((byte *) cemi) + sizeof(CEMI_L_DATA_MESSAGE),
	       &(emi->npdu) + 1, emi->ctrl2 & 0x0f);


    } else {			/* other Message formats are not accepted */


#ifdef DEBUG
	fprintf(stdout, "DEBUG: Not known EMI-Message code %X \n",
		emi->mc);
#endif
	return NULL;		/* if we get something strange return NULL */
    }

    return (byte *) cemi;
}

/**
 * Prepares and sends out the 
 * Tunnelling request. It registers
 * the timeout alarm. The EMI into cEMI
 * message transformation is done using
 * the *convert_EMI_to_CEMI method.
 */
void SendUdp_SendPacket()
{
    EIBNETIP_COMMON_CONNECTION_HEADER *ecch;
    EMI_L_DATA_MESSAGE *emi = (EMI_L_DATA_MESSAGE *) buffer;


    /*
     * *** init tunnelling ack ***
     */

    EIBNETIP_PACKET *packet =
	malloc(sizeof(CEMI_L_DATA_MESSAGE) + HEADER_SIZE_10 +
	       sizeof(EIBNETIP_COMMON_CONNECTION_HEADER) +
	       (emi->ctrl2 & 0x0F));

    packet->head.headersize = HEADER_SIZE_10;
    packet->head.version = EIBNETIP_VERSION_10;
    packet->head.servicetype = htons(TUNNELLING_REQUEST);
    packet->head.totalsize =
	htons(sizeof(CEMI_L_DATA_MESSAGE) + HEADER_SIZE_10 +
	      sizeof(EIBNETIP_COMMON_CONNECTION_HEADER) +
	      (emi->ctrl2 & 0x0F));



    sem_wait(semid);		/* is shared memory available ?? */

    /*
     * *** prepare common connection header ***
     */


    ecch =
	(EIBNETIP_COMMON_CONNECTION_HEADER *) ((byte *) packet +
					       HEADER_SIZE_10);
    ecch->structlength = sizeof(EIBNETIP_COMMON_CONNECTION_HEADER);
    ecch->channelid = active_connection->channelid;
    ecch->sequencecounter = active_connection->sequenceSend;
    ecch->typespecific = 0x00;


    /* 
     * *** prepare cemi message ***
     */


    cemi = (CEMI_L_DATA_MESSAGE *) convert_EMI_to_CEMI(buffer);

    if (cemi != NULL) {		/* it is a valid frame format??? */

	/* copy cemi message into the tunnelling packet */
	memcpy((byte *) packet + HEADER_SIZE_10 +
	       sizeof(EIBNETIP_COMMON_CONNECTION_HEADER), cemi,
	       sizeof(CEMI_L_DATA_MESSAGE) + (emi->ctrl2 & 0x0F));

	/* send packet to data endpoint */
	sendPacket(packet, active_connection->data_endpoint);

	sem_post(semid);	/* free shared memory */

	alarm(TUNNELING_REQUEST_TIME_OUT);	/* set alarm */
    } else {			/* if it is a unknown frameformat */
	sem_post(semid);	/* free shared memory */
    }

    free(packet);
    free(cemi);
}

/**
 * This method blocks until a new
 * message is in the send_udp pipe.
 * If there is an active connection the
 * message is processed, otherwise it is
 * discarded. The message is sent calling
 * the SendUDP_sendPacket method.
 */ 

void readNextFromPipe()
{
    EMI_L_DATA_MESSAGE *emi;

    while (1) {
	if (read(send_udp_pipe,	/* read the first 10 bytes */
		 buffer, sizeof(EMI_L_DATA_MESSAGE)) < 0) {
	    bailOut("Error: can't read from send_pipe");
	}

	emi = (EMI_L_DATA_MESSAGE *) buffer;
	/* length information is in emi->ctrl2&0x0f */

	if (read(send_udp_pipe,
		 buffer + sizeof(EMI_L_DATA_MESSAGE),
		 emi->ctrl2 & 0x0F) < 0) {
	    bailOut("Error: can't read from send_pipe");
	}

	sem_wait(semid);	/* is shared memory available ?? */

	if (active_connection->reset) {	/* no active connection */
	    sem_post(semid);	/* free shared memory */

#ifdef DEBUG
	    fprintf(stdout, "DEBUG: packet discarded\n");
#endif
	    readNextFromPipe();	/* do nothing read next */
	} else {		/* there is a active connection */

	    sem_post(semid);	/* free shared memory */
	    SendUdp_SendPacket();	/* send packet to clients endpoint */
	}

	pause();		/* wait for a signal */

#ifdef DEBUG
	fprintf(stdout, "DEBUG: packet sendet \n");
#endif

	if (ntimeout == 1) {	/* if a timeout has come, wait again, packet shipment is done by the signal handler */

#ifdef DEBUG
	    fprintf(stdout, "DEBUG: packet repeated \n");
#endif
	    pause();
	}
    }
}


/*
 * *********************************************** EOF ***
 */
